#!/usr/bin/perl

#Example of how to use the search api to get information from microbedb
#Requires the user to provide the table name and a hash of column - value pairs to search on

#See object_search_example.pl, if you are searching the regular tables in microbeDB

#This script retrieves all chromosome replicons from a single download version using the object interface
#Then, each replicon is processed
#Finally, the actual replicon sequence is retrieved

use warnings;
use strict;

use lib "/home/mlangill/common/workspace/microbedb/MicrobeDB/";

#we need to use the Search library (this also imports GenomeProject,Replicon, and Gene libs)
use Search;

#Create an object with certain features that we want
#In this case, we are searching for all chromosomes (i.e. not plasmids) from download version 1
my $rep_obj = new Replicon( version_id => '1', rep_type => 'chromosome' );

#Create the search object.
my $search_obj = new Search();

#do the actual search using the replicon object to set the search parameters
#all objects that match the search criteria are returned as an array of the same type of objects
my @result_objs = $search_obj->object_search($rep_obj);

#iterate through each replicon object that was returned
foreach my $curr_rep_obj (@result_objs) {

	#get the name of the replicon
	my $rep_name = $curr_rep_obj->definition;

	#get the priamry identifier for the replicon
	my $rpv_id = $curr_rep_obj->rpv_id;
	
	##############################################################
	#First example of using the table_search method
	
	#Get all chromosomes with a distance that are less than 0.25
	#Note: this uses the table replicon_distance
	
	#Since we are using "less than" instead of the normal equals we need to use the advanced_where search option
	$search_obj->advanced_where(1);
	
	#do the actual table search
	#INPUT: There are two arguements. 1) a scalar as the table name 2) a hash reference containin the table columns and values to search on 
	#Output: an array of hashes. Each record is a hash in the array
	my @distance_results = $search_obj->table_search('replicon_distance',{'rpv_id1' => "= $rpv_id", 'distance' => '< 0.25'});
	
	#loop through each record that was returned
	foreach my $dist_record(@distance_results){
		my $rpv1 = $dist_record->{rpv_id1};
		my $rpv2 = $dist_record->{rpv_id2};
		my $distance = $dist_record->{distance};
		
		print "id1: $rpv1 \t id2: $rpv2 \t dist: $distance \n";	
	}


}

print "All done test!\n";    
