
package Buster;

#Used to find a free node on buster
#perldoc Buster - for more information

use strict;
use warnings;
use Carp;

use Fcntl;
our $AUTOLOAD;


sub new {
	my ( $class, %arg ) = @_;
	my $self = bless {
					_start_node => $arg{start_node} || 4,
					_end_node   => $arg{end_node}   || 45,
					_no_node    => $arg{no_node},
					_log_dir => $arg{log_dir} || '/home/mlangill/logs/',
					_run_nodes => $arg{run_nodes},
	}, $class;

	#Nodes that should never have jobs run on
	#Note: Any special named nodes (e.g. mysql, pathodb1, www1) are not used anyways. Only c0** nodes are used.
	push(
		  @{ $self->{_no_node} },
		  ( 'c001', 'c002', 'c003', 'c042', 'c036' )
	);
	my @run_nodes = $self->get_run_nodes();
	$self->set_run_nodes(\@run_nodes);
    
    #Check for nodes that can not be connected to
    my @bad_nodes = $self->check_nodes();
    foreach(@bad_nodes){
        $self->add_no_node($_);
    }
    
	return $self;
}

sub get_run_nodes{
    my ($self)   = @_;
	my $min_node = $self->get_start_node();
	my $max_node = $self->get_end_node();
	my $log_dir  = $self->get_log_dir();
	my @no_node  = @{ $self->get_no_node() };
	
	my @run_nodes;
	my @nodes = `ruptime -lr`;
	chomp @nodes;

		#last two lines of ruptime is not useful
		pop(@nodes);
		pop(@nodes);

		foreach (@nodes) {

			#Skip nodes that are down
			next if (/down/);

			my @tmp = split;

			#Skip nodes that should have jobs run on them
			my $no_node_flag = 0;
			foreach (@no_node) {
				if ( $tmp[0] =~ /$_/ ) {
					$no_node_flag = 1;
					last;
				}
			}
			next if ($no_node_flag);

			#Don't use nodes with special names (only use e.g  c023)
			next unless($tmp[0] =~ /c0(\d\d)/);

			#Only use nodes within the min and max nodes		       
			next if ( $1 < $min_node or $1 > $max_node );
			
			push(@run_nodes,$tmp[0]);
		}
		return @run_nodes;
}

#Returns a free node on buster the cluster
sub get_node {
	my ($self)   = @_;
	my $min_node = $self->get_start_node();
	my $max_node = $self->get_end_node();
	my $log_dir  = $self->get_log_dir();
	my @no_node  = @{ $self->get_no_node() };

	#Keep looping until a node can be returned
	while (1) {

		my @nodes = `ruptime -lr`;
		chomp @nodes;

		#last two lines of ruptime is not useful
		pop(@nodes);
		pop(@nodes);

		foreach (@nodes) {

			#Skip nodes that are down
			next if (/down/);

			my @tmp = split;

			#Skip nodes that should have jobs run on them
			my $no_node_flag = 0;
			foreach (@no_node) {
				if ( $tmp[0] =~ /$_/ ) {
					$no_node_flag = 1;
					last;
				}
			}
			next if ($no_node_flag);

			#Don't use nodes with special names (only use e.g  c023)
			next unless($tmp[0] =~ /c0(\d\d)/);

			#Only use nodes within the min and max nodes		       
			next if ( $1 < $min_node or $1 > $max_node );

			my $load = $tmp[6];
			chop $load;
			next if ( $load > 1.1 );
			my $time_now  = time;
			my $node_file = $log_dir . $tmp[0];
			if ( -e $node_file ) {

				# make sure the last usage was more than a minute ago!
				my $time = `tail -1 $node_file`;
				chomp $time;
				my $diff = $time_now - $time;
				if ( $diff < 120 ) {
					next;
				}
			}else{
			    #Create the new log file with read and write permissions
			    sysopen(FH, $node_file, O_CREAT)   or die $!;
			    close(FH);
			    `chmod 666 $node_file`;
			}
			my $system_error = `echo $time_now >> $node_file 2>&1`;
			if ($system_error) {
				$system_error =~ s/\n/ /g;
				warn "# WARNING, SYSTEM MESSAGE: $system_error";
			}
			my $time = localtime;

			#Added the following so that only the node number is returned
			#This allows the node to be used with clsh, too!
			$tmp[0] =~ s/c0+//;

			#if($tmp[0] =~ /c0+(\d*)/){
			#    $tmp[0]=$1;
			#}

			return $tmp[0];
		}

		#return 'No node available!';

		#All nodes are busy so wait for a minute and check again
		#carp "WARNING, all nodes are busy: waiting for 1 minute before trying again.";
		sleep(60);

	}
}

#adds a node to the no node list
sub add_no_node{
    my ($self,$node) = @_;
    #my @bad_nodes = @{$self->get_no_node()};
    #push(@bad_nodes,$node);
    #$self->set_no_node(\@bad_nodes);
    push @{$self->{_no_node}}, $node;
    
}

sub check_nodes{
    my ($self)   = @_;
    my @bad_nodes;
	my @run_nodes = $self->get_run_nodes();
	foreach my $curr_node(@run_nodes){
	    my $bad_node=0;
	    my $long_node = $curr_node;
	    $curr_node =~ s/c0+//;
	    
	    #check if the node can be connected to
	    my $sys_msg = `clsh $curr_node ps 2>&1 1>/dev/null`;
	    
	    #put bad nodes on no_node list
	    if($sys_msg =~ /Connection refused/){
	        push(@bad_nodes, $long_node);
	    }
	    
	}
	
	return @bad_nodes;
}

# This takes the place of such accessor definitions as:
#  sub get_attribute { ... }
# and of such mutator definitions as:
#  sub set_attribute { ... }
sub AUTOLOAD {
	my ( $self, $newvalue ) = @_;

	my ( $operation, $attribute ) = ( $AUTOLOAD =~ /(get|set)(_\w+)$/ );

	# Is this a legal method name?
	unless ( $operation && $attribute ) {
		croak "Method name $AUTOLOAD is not in the recognized form (get|set)_
attribute\n";
	}
	unless ( exists $self->{$attribute} ) {
		croak "No such attribute '$attribute' exists in the class ", ref($self);
	}

	# Turn off strict references to enable "magic" AUTOLOAD speedup
	no strict 'refs';

	# AUTOLOAD accessors
	if ( $operation eq 'get' ) {

		# define subroutine
		*{$AUTOLOAD} = sub { shift->{$attribute} };

		# AUTOLOAD mutators
	} elsif ( $operation eq 'set' ) {

		# define subroutine
		*{$AUTOLOAD} = sub { shift->{$attribute} = shift; };

		# set the new attribute value
		$self->{$attribute} = $newvalue;
	}

	# Turn strict references back on
	use strict 'refs';

	# return the attribute value
	return $self->{$attribute};
}

# When an object is no longer being used, this will be automatically called
# and will adjust the count of existing objects
sub DESTROY {
	my ($self) = @_;

}

1;

__END__

=head1 NAME

Buster: Retrieves a free node on buster

=head1 Synopsis

 use Buster;
 
 #Make a default Buster object
 my $buster_obj = new Buster();

 #Make a Buster object with specific beginning and end nodes 
 my $buster_obj = new Buster(start_node => 10, end_node => 20);

 #Specify your own temp file
 my $buster_obj = new Buster(log_dir =>  '/home/mlangill/research/logs/');


 #Specify nodes not to use (some nodes never have jobs run on them such as mysql,pathodb1, etc)
 #Note: must be passed as an array reference
 my $buster_obj = new Buster(no_node =>[c007,c005]);

 #Retrieve a free node on buster
 #If all nodes are busy, it will sleep for 1 minute before trying again 
 my $node = $buster_obj->get_node();

=head1 AUTHOR

Morgan Langille

=head1 Date

May 11th, 2006

=cut


