
package Util;
#Contains common subroutines that can be used in various circumstances

use strict;
use warnings;
use Carp;

sub new{
    return bless {},shift;
}

#Takes a filename that may include a path too it and returns only the filename with no extension
sub path2filename{
    my ($self,$filename)=@_;
    if( $filename =~ /(.*\/)*(\w+)(\.\w+)?/){
	return $2;
    }else{
	die "ERROR: Can't extract filename from path: $!";
	return;
    }

}

#sums an array
sub sum{
    my ($self,@numbers)=@_;
    my $sum;
    foreach(@numbers){
	$sum+=$_;
    }
    return $sum;
}

#Averages an array
sub avg{
    my($self,@numbers)=@_;
    return sum($self,@numbers)/scalar(@numbers);
}

#Returns the smallest value in the array
sub min{
    my($self,@numbers)=@_;
    @numbers=sort{$a <=> $b} @numbers;
    return shift(@numbers);
}

#Returns the largest value in the array
sub max{
    my($self,@numbers)=@_;
    @numbers=sort{$a <=> $b} @numbers;
    return pop(@numbers);
}

#Rounds a number
#Last arguement is the number of decimal places to round to
sub round{
    my ($self,$number,$decimal_places)=@_;
    my $mult_factor= 10 ** $decimal_places;
    $number *= $mult_factor;
    #round the number (can be negative)
    return int($number + .5 * ($number <=> 0))/$mult_factor;
   
}


=head1 Util

Util: Contains numerous methods that can be used for common needs

=head1 Synopsis

use Util;

my @numbers = qw(4 5 6 10);

my $util_obj = new Util();

print "Sum is ", $util_obj->sum(@numbers);

print "Average is ", $util_obj->average(@numbers);

print "Max is ", $util_obj->max(@numbers);

print "Min is ", $util_obj->min(@numbers);

=head1 AUTHOR

Morgan Langille

=head1 Date

May 10th, 2006

=cut



1;
